import time

import numpy as np

from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceLight(Device):

    def __init__(self, device_id=None, config=None):
        super(DeviceLight, self).__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_INTENSITY, readonly=False, default_type=np.float16, min_value=0, max_value=100,
                              unit="%", decimals=1, set_function=self.set_intensity, set_value_when_set_function=False)
        self.create_attribute(ATTR_SETPOINT, default_value=75, readonly=False, default_type=np.float16, min_value=0,
                              max_value=100, unit="%", decimals=1)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def set_intensity(self, value, callback=None):
        raise NotImplementedError()

    @expose_method()
    def turn_on(self, callback=None):
        self.logger.info(u"Turn On")
        self.set_intensity(self.get_value(ATTR_SETPOINT))

    @expose_method()
    def turn_off(self, callback=None):
        self.logger.info(u"Turn Off")
        self.set_intensity(0)

    @expose_method()
    def stop(self):
        self.logger.info(u"Stop device")
        self.turn_off()

class DeviceLightAttribute(DeviceLight):

    def __init__(self, device, attribute, output_switch_attribute=None, device_id=None, config=None):
        self.device = device
        self.attribute = Attribute.list_attribute(attribute)
        self.output_switch_attribute = output_switch_attribute
        if output_switch_attribute is not None:
            self.output_switch_attribute = Attribute.list_attribute(output_switch_attribute)
        super().__init__(device_id, config)

    def connect(self, *args):
        super(DeviceLightAttribute, self).connect(*args)
        self.device.attach_attribute_callback(self.attribute, self.value_changed, key_filter=VALUE)
        self.device.attach_attribute_callback(ATTR_STATUS, self.set_status, key_filter=VALUE)

    def _init_attributes(self):
        DeviceLight._init_attributes(self)
        if self.output_switch_attribute is not None:
            self.add_attribute(ATTR_OUTPUT_SWITCH, self.device.get_attribute(self.output_switch_attribute))

    def value_changed(self, value):
        if value is None:
            return
        value_min, value_max = self.device.get_attribute(self.attribute + [MIN]), self.device.get_attribute(
            self.attribute + [MAX])
        intensity = ((value - value_min) * 100) / (value_max - value_min)
        self.set_raw_value(ATTR_INTENSITY, intensity)

    def set_intensity(self, value, callback=None):
        self.logger.info(u"Set light intensity to {} %".format(value))
        if value > 0:
            self.set_value(ATTR_SETPOINT, value)
        value_min, value_max = self.device.get_attribute(self.attribute + [MIN]), self.device.get_attribute(
            self.attribute + [MAX])
        attribute_value = ((value * (value_max - value_min) / 100) + value_min)
        self.device.set_attribute(self.attribute + [VALUE], attribute_value)

    @expose_method()
    def stop(self):
        self.device.stop()

    def close(self):
        self.device.detach_attribute_callback(self.attribute, self.value_changed)
        self.device.detach_attribute_callback(ATTR_STATUS, self.set_status)
        return super(DeviceLightAttribute, self).close()

    @expose_method()
    def turn_on(self, callback=None):
        self.logger.info(u"Turn On")
        if self.output_switch_attribute is not None:
            self.device.set_value(self.output_switch_attribute, True)
        self.set_intensity(self.get_value(ATTR_SETPOINT))

    @expose_method()
    def turn_off(self, callback=None):
        self.logger.info(u"Turn Off")
        if self.output_switch_attribute is not None:
            self.device.set_value(self.output_switch_attribute, False)
        self.set_intensity(self.get_value(ATTR_SETPOINT))


import re
import time

import numpy as np
import serial

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceZled(DeviceSocket):
    error_pattern = re.compile('^F00[1-3]$')
    terminator = b"\r"
    push_commands_max = 1
    error_statuses = {1: "LED Temperature sensor", 2: "Undefined", 3: "LED Connection", 4: "LED overtemperature",
                      5: "PCB overtemperature", 6: "Fan", 7: "Undefined"}

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command("GN\r")
        self.command("G#\r")
        self.command("GO\r")
        self.command("GV\r", callback=_finish_configuration)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_SOFTWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_MODEL, readonly=True)
        self.create_attribute(ATTR_INTENSITY, unit="%", default_type=np.int, min_value=0, max_value=100, readonly=False,
                              set_function=self.set_intensity)
        self.create_attribute(ATTR_OUTPUT_SWITCH, readonly=True, default_type=bool)
        self.create_attribute(ATTR_SETPOINT, default_value=75, readonly=False, default_type=np.float16, min_value=0,
                              max_value=100, unit="%", decimals=1)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DevicePort.handle_readout(self, readout_buffer)
        is_error_message = self.error_pattern.match(output)
        if is_error_message is not None:
            self.set_status(STATUS_ERROR)
            self.handle_command_error(command, readout_buffer)
        elif command == "GS\r":
            status_code = int("0x{}".format(output[2:]), 0)
            if status_code == 0:
                # if self.get_value(ATTR_OUTPUT_SWITCH):
                #     self.set_status(STATUS_BUSY)
                # else:
                self.set_status(STATUS_IDLE)
            else:
                self.set_status(STATUS_ERROR)
                self.handle_command_error(command, self.error_statuses[status_code])
        elif command == "GI\r":
            self.set_raw_value(ATTR_INTENSITY, int(output[2:]))
        elif command == "GO\r":
            self.set_raw_value(ATTR_OUTPUT_SWITCH, int(output[2:]))
        elif command == "GV\r":
            self.set_raw_value(ATTR_SOFTWARE_VERSION, output[2:])
        elif command == "GN\r":
            self.set_raw_value(ATTR_MODEL, output[2:])
        elif command == "G#\r":
            self.set_raw_value(ATTR_SERIAL_NUMBER, output[2:])

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [str(command), str(readout_buffer)], token)

    def start_polling(self):
        DeviceSocket.start_polling(self)
        # Poll status
        self.poll_command("GS\r", 1000)
        # Poll intensity
        self.poll_command("GI\r", 1000)
        # Poll output state
        self.poll_command("GO\r", 1000)

    def set_intensity(self, value, callback=None):
        self.logger.info(u"Set intensity to {}%".format(value))
        if value > 0:
            self.set_value(ATTR_SETPOINT, value)
        return self.command("SI{:03}\r".format(int(value)), callback, with_token=True)

    @expose_method()
    def turn_on(self, callback=None):
        self.logger.info(u"Turn On")
        self.set_intensity(self.get_value(ATTR_SETPOINT))
        return self.command("SO1\r", callback, with_token=True)

    @expose_method()
    def turn_off(self, callback=None):
        self.logger.info(u"Turn Off")
        self.set_intensity(0)
        return self.command("SO0\r", callback, with_token=True)

    @expose_method()
    def stop(self):
        self.logger.info(u"Stop device")
        self.turn_off()

class DeviceZledSocket(DevicePort, DeviceZled):
    terminator = b"\r"

    def __init__(self, port, baud_rate=9600, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        DevicePort.__init__(self, port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b"GS\r")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            output = self.read_all().strip()
            is_error_message = self.error_pattern.match(output)
            return is_error_message is None
        return False



import oyaml as yaml
from PyQt5.QtCore import QLocale
from oyaml import MappingNode
from pint import UnitRegistry
from pint.quantity import build_quantity_class
import logging
QLocale.setDefault(QLocale("US"))


# TODO: Make save positions group resizable
# TODO: Offer tooltip for saved comment so long comment can be seen
# TODO: Add process ID for session device
# TODO: Somehow make sure, that not two instances of session are running

class GuiTypeError(Exception):
    pass


class WriteException(Exception):
    pass


class DeviceError(Exception):
    pass


class DeviceUnitError(DeviceError):
    pass


class CommandFormatException(DeviceError):
    pass


class DeviceServerError(DeviceError):
    pass


class DeviceClientError(DeviceError):
    pass


class DeviceUnknownError(DeviceError):
    pass


class MacroException(DeviceError):
    pass


class StepSetpointException(MacroException):
    pass


class StepTimeoutException(MacroException):
    pass


def device_constructor(loader, suffix, node):
    arguments = loader.construct_mapping(node, deep=True)
    mapping_node = MappingNode(u'tag:yaml.org,2002:python/object{}'.format(suffix), [])
    new_device = loader.construct_object(mapping_node)
    new_device.__init__(**arguments)
    return new_device


yaml.add_multi_constructor(u'!Device', device_constructor, Loader=yaml.Loader)
yaml.add_multi_constructor(u'!Widget', device_constructor, Loader=yaml.Loader)

# Logging setup
SEP = "@!"  # Separator character

"""
Section of units setup, helper functions, ...
"""
units = UnitRegistry()
units.derived_units = []
q_class = build_quantity_class(units)
session = None


def pint_constructor(loader, node):
    value = loader.construct_scalar(node)
    return units.Quantity(value)


def pint_dumper(dumper, data):
    return dumper.represent_scalar('!Pint', str(data))


yaml.add_representer(units.Quantity, pint_dumper)
yaml.add_constructor(u'!Pint', pint_constructor)

logging.MACRO_STATE = 51
logging.addLevelName(logging.MACRO_STATE, "MACRO_STATE")


from collections import deque
from time import perf_counter_ns, sleep

import zmq

from kamzik3.constants import WORKER_READY, MSG_TERMINATE, MSG_EMPTY


class Worker:
    running = False

    def __init__(self, id, master_host, master_port, zmq_context=None):
        self.instructions_set = []
        self.zmq_context = zmq_context
        self.master_host = master_host
        self.master_port = master_port
        self.id = id.encode()

    def _init_socket(self):
        if self.zmq_context is None:
            self.zmq_context = zmq.Context.instance()

        self.socket_worker = self.zmq_context.socket(zmq.REQ)
        self.socket_worker.setsockopt(zmq.IDENTITY, self.id)
        self.socket_worker.connect("tcp://{}:{}".format(self.master_host, self.master_port))

    def run(self):
        self.running = True
        self._init_socket()
        self.socket_worker.send_multipart([WORKER_READY] + self.instructions_set)
        while self.running:
            request = self.socket_worker.recv_multipart()
            if request[2] == MSG_TERMINATE:
                self.stop()
                continue
            self.socket_worker.send_multipart(request[:2] + self.get_payload(request[2:]))
        self.socket_worker.close()

    def get_payload(self, request):
        raise NotImplementedError()

    def stop(self):
        self.running = False


class BalancedServer:
    running = False

    def __init__(self, host, port, workers_count=5, worker_class=Worker):
        self.zmq_context = zmq.Context.instance()
        self.host = host
        self.port = port
        self.socket_frontend = None
        self.socket_backend = None
        self.port_backend = None
        self.workers_count = workers_count
        self.worker_class = worker_class

    def _init_frontend(self):
        self.socket_frontend = self.zmq_context.socket(zmq.ROUTER)
        self.socket_frontend.setsockopt(zmq.IDENTITY, "FE@{}:{}".format(self.host, self.port).encode())
        self.socket_frontend.bind("tcp://{}:{}".format(self.host, self.port))

    def _init_backend(self):
        self.socket_backend = self.zmq_context.socket(zmq.ROUTER)
        self.socket_backend.setsockopt(zmq.IDENTITY, "BE@{}:{}".format(self.host, self.port).encode())
        self.port_backend = self.socket_backend.bind_to_random_port("tcp://{}".format(self.host))

    def _init_workers(self):
        for i in range(self.workers_count):
            worker = self.worker_class("W_{}".format(i), self.host, self.port_backend, self.zmq_context)
            self.local_workers.append(worker)
            worker.start()

    def run(self):
        self.running = True
        self.registered_workers = {}
        self.local_workers = []
        self._init_frontend()
        self._init_backend()
        self.poller = zmq.Poller()
        self.poller.register(self.socket_backend, zmq.POLLIN)
        self.idle_workers = []
        self._init_workers()
        self.main_loop()

        self.socket_backend.close()
        self.socket_frontend.close()

    def main_loop(self):
        while self.running:
            sockets = dict(self.poller.poll(timeout=10))

            if self.socket_backend in sockets:
                # Handle worker activity on the backend
                request = self.socket_backend.recv_multipart()
                worker, _, client = request[:3]
                if not self.idle_workers:
                    # Poll for clients now that a worker is available
                    self.poller.register(self.socket_frontend, zmq.POLLIN)
                self.idle_workers.append(worker)

                if client != WORKER_READY:
                    # If client reply, send rest back to frontend
                    self.socket_frontend.send_multipart([client] + request[3:])
                else:
                    self.registered_workers[worker] = request[3:]

            if self.socket_frontend in sockets:
                # Get next client request, route to last-used worker
                request = self.socket_frontend.recv_multipart()

                worker = self.idle_workers.pop(0)
                instruction = request[2]
                self.socket_backend.send_multipart([worker, MSG_EMPTY] + request)
                if not self.idle_workers:
                    # Don't poll clients if no workers are available
                    self.poller.unregister(self.socket_frontend)

    def stop(self):
        for worker in self.local_workers:
            self.socket_backend.send_multipart([worker.id, MSG_EMPTY, MSG_EMPTY, MSG_EMPTY, MSG_TERMINATE])

        self.local_workers = []
        self.running = False

    def backend_address(self):
        return self.host, self.port_backend


class Publisher():
    running = False

    def __init__(self, host, port, zmq_context=None):
        self.zmq_context = zmq_context
        if self.zmq_context is None:
            self.zmq_context = zmq.Context.instance()
        self.host = host
        self.port = port
        self.publisher_events = {}
        self.messages_queue = deque()
        self._init_socket()

    def _init_socket(self):
        self.socket_publisher = zmq.Context.instance().socket(zmq.PUB)
        self.socket_publisher.setsockopt(zmq.IDENTITY, b"P_0")
        self.socket_publisher.bind("tcp://{}:{}".format(self.host, self.port))

    def run(self):
        timeout = 0.005
        self.running = True
        while self.running:
            publish_time = perf_counter_ns()
            while self.messages_queue:
                header, message, min_timeout = self.messages_queue.popleft()
                last_message_event = self.publisher_events.get(header, None)
                if (last_message_event is None) or (publish_time - last_message_event[0]) >= min_timeout:
                    self._publish_message(header, message)
                    self.publisher_events[header] = [publish_time, [header, None, min_timeout]]
                else:
                    last_message_event[1][1] = message

            for event in list(self.publisher_events.values()):
                if (publish_time - event[0]) >= event[1][2]:
                    if event[1][1] is not None:
                        self._publish_message(event[1][0], event[1][1])
                        self.publisher_events[event[1][0]] = [publish_time, [event[1][0], None, event[1][2]]]
                    else:
                        del self.publisher_events[event[1][0]]

            elapsed_ns = (perf_counter_ns() - publish_time) * 1e-9
            sleep_timeout = timeout - elapsed_ns
            sleep(sleep_timeout if sleep_timeout > 0 else 0)

    def _publish_message(self, header, message):
        raise NotImplementedError()

    def push_message(self, header, message, token=None, min_timeout=100e6):
        try:
            if token is not None:
                header += u"." + str(token)
            self.messages_queue.append((header, message, min_timeout))
        except TypeError:
            self.logger.exception(u"Error pushing message to client")

    def stop(self):
        self.running = False



def device_id_generator():
    """
    Generate ID for unnamed device.
    :return: string
    """
    i = 0
    while True:
        yield u"Device_{}".format(i)
        i += 1


def token_generator():
    """
    Generate Token number up to 2**8.
    This token is used for client server communication.
    :return: int
    """
    token, max_token = 1, pow(2, 8)
    while True:
        yield token
        token += 1
        if token >= max_token:
            token = 1


def axis_name_generator(range_of_axis=1):
    """
    Axis name generator for n axis scan.
    @type range_of_axis: int
    """
    current = 1
    names = [u"Slow", u"Slower", u"Slowest", u"Fast", u"Faster", u"Fastest"]
    mapping = [[3], [0, 3], [1, 0, 3], [1, 0, 3, 4], [2, 1, 0, 3, 4], [2, 1, 0, 3, 4, 5]]
    col = range_of_axis - 1

    while current <= range_of_axis:
        if range_of_axis <= 6:
            yield u"{}".format(names[mapping[col][current - 1]])
        elif current == range_of_axis:
            yield u"Fast {}".format(current)
        elif current == 1:
            yield u"Slow {}".format(current)
        else:
            yield u"{}".format(current)
        current += 1



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kamzik3",
    version="0.7",
    author="Martin Domaracky",
    author_email="sc@cfel.de",
    description="Device controlling framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.desy.de/cfel-sc/kamzik3",
    packages=setuptools.find_packages(),
    package_data={'kamzik3': ['*.yml'], 'kamzik3.example': ['*.yml', "*.att"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
    ],
    install_requires=[
        "numpy>=1.16",
        "pyzmq>=18.0",
        "pint>=0.16.0",
        "bidict>=0.18",
        "pyqt5>=5.15.0",
        "pyqtgraph>=0.11.0",
        "pyserial>=3.4",
        "oyaml>=0.9",
        "psutil>=5.6.0",
        "natsort>=7.0.1",
        "reportlab>=3.5.42",
        "pandas>=1.0.4"
        "tables>=3.6.1"
    ],
    extras_requires={
        "pytango": [],
        "pyopengl": [],
        "sysutil": [],
        "pydaqmx": [],
        "pypiwin32": [],
        "rocketchat-API": [],
    },
    python_requires='>=3.7',
)



import logging
import os
import shutil
import time
from collections import OrderedDict
from threading import RLock

import numpy as np
import oyaml as yaml
import psutil

import kamzik3
from kamzik3 import DeviceUnknownError, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attributeLogger import AttributeLogger, ATTR_LOGGING
from kamzik3.devices.device import Device
from kamzik3.devices.deviceClient import DeviceClient
from kamzik3.devices.general.measurementGroup import MeasurementGroup
from kamzik3.snippets.snippetLogging import set_rotating_file_handler, set_sys_exception_handler, get_console_handler, \
    set_rocket_chat_handler
from kamzik3.snippets.snippetsControlLoops import control_asyncore_loop, control_device_poller, control_port_read_loop, \
    control_device_connection_poller
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsTimer import CallbackTimer
from kamzik3.snippets.snippetsYaml import YamlDumper, YamlConfigSerializer
from multiprocessing import Process

class DeviceSession(Device):
    LOG_DEVICE_OUTPUT_FILE = u"devices.log"
    LOG_CONSOLE_OUTPUT_FILE = u"console.log"
    LOG_ATTRIBUTES_OUTPUT_FILE = u"attributes.log"
    LOG_GUI_OUTPUT_FILE = u"gui.log"
    LOG_MACRO_OUTPUT_FILE = u"macro.log"
    log_level = logging.INFO
    config_lock = RLock()

    def __init__(self, device_id=None, config=None):
        kamzik3.session = self
        self.devices = OrderedDict()
        if config is None:
            self.server = None
            self.publisher = None
        else:
            self.server = config.get("server")
            self.publisher = config.get("publisher")
        self.measurement_groups = {}
        self.polling_timer = CallbackTimer(1000, self.measure_cpu_ram)
        self.os_process = psutil.Process(os.getpid())
        self.session_config = None
        super(DeviceSession, self).__init__(device_id, config)
        self.connect()

        if self.server is not None:
            self.server.start()
        if self.publisher is not None:
            self.publisher.start()

    def _init_attributes(self):
        super(DeviceSession, self)._init_attributes()
        self.create_attribute(ATTR_LOG_DIRECTORY, set_function=self.set_log_directory, readonly=True)
        self.create_attribute(ATTR_RESOURCE_DIRECTORY, set_function=self.set_resource_directory, readonly=True)
        self.create_attribute(ATTR_CACHE_DIRECTORY, set_function=self.set_cache_directory, readonly=True)
        self.create_attribute(ATTR_DEVICES_COUNT, default_value=0, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_ALLOW_CONSOLE_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_GUI_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_DEVICE_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_ATTRIBUTE_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_MACRO_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_RC, group=GROUP_RC, default_value=False, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MAIN_ROOM, group=GROUP_RC, default_value=self.device_id.lower(), readonly=True)
        self.create_attribute(ATTR_AUTH_TOKEN, group=GROUP_RC, default_value=None, readonly=True, display=False)
        self.create_attribute(ATTR_USER_NAME, group=GROUP_RC, default_value=None, readonly=True, display=False)
        self.create_attribute(ATTR_PASSWORD, group=GROUP_RC, default_value=None, readonly=True, display=False)
        self.create_attribute(ATTR_RC_MACRO_DISCUSSION, group=GROUP_RC, default_value=None, readonly=True)
        self.create_attribute(ATTR_HOST, group=GROUP_RC, default_value="https://cfel-rocketchat.desy.de", readonly=True)
        self.create_attribute(ATTR_LOOPS_RUNNING, default_value=False, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_PROCESS_ID, default_value=os.getpid(), readonly=True)
        self.create_attribute(ATTR_CPU_USAGE, default_value=0, default_type=np.float, readonly=True, unit="%",
                              decimals=4)
        self.create_attribute(ATTR_MEMORY_USAGE, default_value=0, default_type=np.float, readonly=True, decimals=3,
                              unit="MB", factor=10e-7)
        self.create_attribute(ATTR_AVAILABLE_MEMORY, default_value=0, default_type=np.float,
                              readonly=True, unit="MB", factor=10e-7, decimals=3)
        self.set_raw_value(ATTR_AVAILABLE_MEMORY, psutil.virtual_memory().total)

    def add_measurement_group(self, measurement_group):
        assert isinstance(measurement_group, MeasurementGroup)
        self.measurement_groups[measurement_group.group_id] = measurement_group

    def set_session(self, session):
        self.devices[self.device_id] = self
        self.session = self
        self.set_value(ATTR_DEVICES_COUNT, len(self.devices))

    def measure_cpu_ram(self):
        self.set_raw_value(ATTR_CPU_USAGE, self.os_process.cpu_percent())
        self.set_raw_value(ATTR_MEMORY_USAGE, self.os_process.memory_info().rss)

    def set_config(self, config_file):
        with self.config_lock:
            with open(config_file, "r") as configFile:
                self.session_config = (config_file, yaml.load(configFile, Loader=YamlConfigSerializer))

    def update_config(self, device_id, attribute, value):
        with self.config_lock:
            if self.session_config is None:
                return
            else:
                config_file, parsed_config = self.session_config
                device_config = parsed_config["devices"][device_id]
                if "config" not in device_config.conf:
                    device_config.conf["config"] = {"attributes": {tuple(attribute): value}}
                elif "attributes" not in device_config.conf["config"]:
                    device_config.conf["config"]["attributes"] = {tuple(attribute): value}
                else:
                    device_config.conf["config"]["attributes"][tuple(attribute)] = value
                with open(config_file, "w") as fp:
                    yaml.dump(parsed_config, fp, Dumper=YamlDumper)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()

            if self.get_value(ATTR_ALLOW_CONSOLE_LOG):
                self.set_console_log_directory(u"Console")
            if self.get_value(ATTR_ALLOW_GUI_LOG):
                self.set_gui_log_directory(u"Gui")
            if self.get_value(ATTR_ALLOW_DEVICE_LOG):
                self.set_device_log_directory(u"Device")
            if self.get_value(ATTR_ALLOW_ATTRIBUTE_LOG):
                self.set_attribute_log_directory(u"Attributes")
                self._config_attributes_log()
            if self.get_value(ATTR_ALLOW_MACRO_LOG):
                self.set_macro_log_directory(u"Macro")
            if self.get_value(ATTR_CACHE_DIRECTORY) is None:
                self.set_value(ATTR_CACHE_DIRECTORY, ".cache")

            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def _config_attributes_log(self):
        if self.config:
            logger = self.get_device("AttributeLogger")
            for device_id, attribute in self.config.get("logged_attributes", {}).items():
                if isinstance(attribute, list):
                    for sub_attribute in attribute:
                        logger.add_logged_attribute(device_id, sub_attribute)
                else:
                    logger.add_logged_attribute(device_id, attribute)
            logger.set_attribute((ATTR_LOGGING, VALUE), True)

    def start_polling(self):
        self.polling_timer.start()

    def stop_polling(self):
        self.polling_timer.stop()

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def get_device(self, device_id):
        try:
            return self.devices[device_id]
        except KeyError:
            master_server, master_port = self.config.get("master_server", False), self.config.get("master_port", False)
            if master_server and master_port:
                return DeviceClient(master_server, master_port, device_id)
            raise DeviceUnknownError("Device {} was not found.".format(device_id))

    def change_device_id(self, device_id, new_device_id):
        device = self.get_device(device_id)
        device.device_id = new_device_id
        self.devices[new_device_id] = self.devices.pop(device_id)

    def add_device(self, device):
        assert isinstance(device, Device)

        if device.device_id not in self.devices:
            self.devices[device.device_id] = device
            device.session = self
            self.set_value(ATTR_DEVICES_COUNT, len(self.devices))
        else:
            raise DeviceError("Device {} is already registered within session.".format(device.device_id))

    def remove_device(self, device):
        assert isinstance(device, Device)
        if device.device_id in self.devices:
            device.session = None
            del self.devices[device.device_id]
            self.set_value(ATTR_DEVICES_COUNT, len(self.devices))

    def get_devices(self, class_filter=None, attribute_filter=None, method_filter=None):
        """
        Filter devices in current session
        :param class_filter: list of classes
        :param attribute_filter: attribute
        :param method_filter: method name
        :return: list of filtered devices
        """
        filtered_devices = self.devices.copy()
        # Filter over device class
        if class_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                if class_filter in device.qualified_name:
                    del filtered_devices[device_id]
        # Filter over device attributes
        if attribute_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                if device.get_attribute(attribute_filter) is None:
                    del filtered_devices[device_id]
        # Filter over device methods
        if method_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                filtered_out_methods = filter(lambda val: val[0] == method_filter, device.exposed_methods)
                try:
                    next(filtered_out_methods)
                except StopIteration:
                    del filtered_devices[device_id]

        return filtered_devices

    def set_log_directory(self, value):
        self.logger.info(u"Setting log directory to: {}".format(value))
        if not os.path.exists(value):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(value))
            os.makedirs(value)

    def set_resource_directory(self, value):
        self.logger.info(u"Setting resource directory to: {}".format(value))
        if not os.path.exists(value):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(value))
            os.makedirs(value)

    def set_cache_directory(self, value):
        self.logger.info(u"Setting cache directory to: {}".format(value))
        if not os.path.exists(value):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(value))
            os.makedirs(value)

    def set_console_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set console log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)
        self.logger.info(u"Setting console log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Console")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_CONSOLE_OUTPUT_FILE))
        set_sys_exception_handler(logger)
        # if self.get_value(ATTR_ALLOW_RC):
        #     discussion = "{}-{}".format(self.device_id.lower(), "error-log")
        #     set_rocket_chat_handler(logger, level=logging.ERROR, main_room=self.device_id.lower(), discussion=discussion)
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

    def set_gui_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set gui log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting gui log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Gui")
        # if self.get_value(ATTR_ALLOW_RC):
        #     discussion = "{}-{}".format(self.device_id.lower(), "error-log")
        #     set_rocket_chat_handler(logger, level=logging.ERROR, main_room=self.device_id.lower(), discussion=discussion)
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_GUI_OUTPUT_FILE))
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def set_attribute_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set device log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting device log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        logger = AttributeLogger(os.path.join(log_directory, self.LOG_ATTRIBUTES_OUTPUT_FILE),
                                 device_id="AttributeLogger")
        self.attributes.update({
            "Attribute logger": logger.attributes
        })

    def set_device_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set device log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting device log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Device")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_DEVICE_OUTPUT_FILE))
        # if self.get_value(ATTR_ALLOW_RC):
        #     discussion = "{}-{}".format(self.device_id.lower(), "error-log")
        #     set_rocket_chat_handler(logger, level=logging.ERROR, channel=channel)
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

    def set_macro_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set gui log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting gui log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Macro")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_MACRO_OUTPUT_FILE))
        if self.get_value([GROUP_RC, ATTR_ALLOW_RC]):
            discussion_name = self.get_value([GROUP_RC, ATTR_RC_MACRO_DISCUSSION])
            if discussion_name is None:
                discussion_name = "{}-{}".format(self.device_id.lower(), "macro-log")
                self.set_value([GROUP_RC, ATTR_RC_MACRO_DISCUSSION], discussion_name)
            user_credentials, auth_token = self._get_rc_auth()
            set_rocket_chat_handler(logger, level=logging.MACRO_STATE, user_credentials=user_credentials,
                                    auth_token=auth_token, main_room=self.device_id.lower(), discussion=discussion_name)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def _get_rc_auth(self):
        rc_user = self.get_value([GROUP_RC, ATTR_USER_NAME])
        rc_pass = self.get_value([GROUP_RC, ATTR_PASSWORD])
        rc_auth_token = self.get_value([GROUP_RC, ATTR_AUTH_TOKEN])
        return (rc_user, rc_pass), rc_auth_token

    def start_control_loops(self):
        if not self.get_value(ATTR_LOOPS_RUNNING):
            self.start_asyncore_loop()
            self.start_device_connection_poller()
            self.start_device_poller_loop()
            self.start_port_read_loop()
            self.set_value(ATTR_LOOPS_RUNNING, True)

    def stop_control_loops(self):
        if self.get_value(ATTR_LOOPS_RUNNING):
            self.stop_asyncore_loop()
            self.stop_port_read_loop()
            self.stop_device_poller_loop()
            self.stop_device_connection_poller()
            self.set_value(ATTR_LOOPS_RUNNING, False)

        self.logger.info("Closing all control loops")
        # Here we try to join all control loops giving them 1 second timeout
        control_asyncore_loop.join(1)
        control_device_poller.join(1)
        control_port_read_loop.join(1)
        control_device_connection_poller.join(1)

        self.logger.info("Control loops closed")

    def stop(self):
        """
        Stop and disconnect all devices in Session
        :return: None
        """
        self.stop_polling()

        if self.server is not None:
            self.server.stop()

        if self.publisher is not None:
            self.publisher.stop()

        self.session = None
        del self.devices[self.device_id]

        for device in list(self.devices.values()):
            device.disconnect()

        if self.get_value(ATTR_LOOPS_RUNNING):
            self.stop_control_loops()
            self.set_value(ATTR_LOOPS_RUNNING, False)

    @staticmethod
    def start_asyncore_loop():
        control_asyncore_loop.start()

    @staticmethod
    def stop_asyncore_loop():
        control_asyncore_loop.stop()

    @staticmethod
    def start_device_poller_loop():
        control_device_poller.start()

    @staticmethod
    def stop_device_poller_loop():
        control_device_poller.stop()

    @staticmethod
    def start_port_read_loop():
        control_port_read_loop.start()

    @staticmethod
    def stop_port_read_loop():
        control_port_read_loop.stop()

    @staticmethod
    def start_device_connection_poller():
        control_device_connection_poller.start()

    @staticmethod
    def stop_device_connection_poller():
        control_device_connection_poller.stop()

    def set_high_process_priority(self):
        """
        Set high process affinity.
        In Windows use win32process.REALTIME_PRIORITY_CLASS, for Linux change nice to -20.
        :return: None
        """
        if psutil.WINDOWS:
            try:
                import win32api, win32process, win32con
            except ImportError:
                self.logger.error("Package pypiwin32 is not installed")
            pid = win32api.GetCurrentProcessId()
            handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
            win32process.SetPriorityClass(handle, win32process.REALTIME_PRIORITY_CLASS)
            # win32process.SetThreadPriority(win32api.GetCurrentThread(), win32process.THREAD_PRIORITY_TIME_CRITICAL)
        else:
            try:
                process = psutil.Process(os.getpid())
                process.nice(-20)
            except psutil.AccessDenied as e:
                self.logger.error("Could not raise process priority")

    @expose_method()
    def clear_cache_folder(self):
        """
        Clear all files from current '.cache' folder.
        :return: None
        """
        folder = self.get_value(ATTR_CACHE_DIRECTORY)
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                self.logger.error('Failed to delete %s. Reason: %s' % (file_path, e))

def server_process(conf, stop_event):
    def _process(conf, stop_event):
        import kamzik3
        import yaml
        from time import sleep
        with open(conf, "r") as configFile:
            yaml.load(configFile, Loader=yaml.Loader)

        # Start control loop for devices specified in configuration file
        kamzik3.session.set_config(conf)
        kamzik3.session.start_control_loops()
        while not stop_event.is_set():
            sleep(1)
        kamzik3.session.stop()

    return Process(target=_process, args=(conf, stop_event))



from PyQt5.QtWidgets import QSizePolicy, QComboBox

from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget


class AttributeComboxWidget(AttributeWidget):

    def _set_input_widget(self):
        self.input_widget = QComboBox()
        self.input_widget.addItems(["None"] + [str(i) for i in self.attribute[TYPE_LIST_VALUES]])
        self.input_widget.model().item(0).setEnabled(False)
        if self.attribute[READONLY]:
            self.input_widget.setDisabled(True)

        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

    def set_value(self, value):
        if value is None:
            return

        value = str(value)
        index = self.input_widget.findText(value)
        if value is None or index == -1:
            self.input_widget.setCurrentIndex(0)
        else:
            self.input_widget.setCurrentIndex(index)

    def get_attribute_value(self):
        return self.input_widget.currentText()

    def get_widget_value(self):
        return self.input_widget.currentText()



import re
import time
from threading import Thread, Event

import numpy as np
from pint.errors import UndefinedUnitError

import kamzik3
from kamzik3 import DeviceError, DeviceUnitError, units
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

""" Example of yaml configuration
Lexium0: &Lexium0 !Device:kamzik3.devices.deviceLexium.DeviceLexium
    device_id: Lexium0
    host: 192.168.83.60
Lexium0_axis_0: &Lexium0_axis_0 !Device:kamzik3.devices.deviceLexium.DeviceLexiumChannel
    device: *Lexium0
    channel: 0
    device_id: MotorAxis0
    config:
        attributes:
            !!python/tuple [Encoder enabled, Value]: 1
            !!python/tuple [Rotation direction, Value]: 0
            !!python/tuple [Use mark index, Value]: 0
            !!python/tuple [Position, Factor]: 2.0833e-6
            !!python/tuple [Position, Unit]: deg
            !!python/tuple [Position +limit, Factor]: 2.0833e-6
            !!python/tuple [Position +limit, Unit]: deg
            !!python/tuple [Position -limit, Factor]: 2.0833e-6
            !!python/tuple [Position -limit, Unit]: deg
            !!python/tuple [Backlash distance, Unit]: deg
            !!python/tuple [Backlash distance, Value]: 1 deg
            !!python/tuple [Maximum velocity, Value]: 80000
            !!python/tuple [Microstep resolution, Value]: 256
            !!python/tuple [Run current, Value]: 50%
            !!python/tuple [Stall factor, Value]: 30
            !!python/tuple [Position, Tolerance]: [0.1 deg, 0.1 deg]
"""


class DeviceLexium(DeviceSocket):
    terminator = b">"
    response_timeout = 3000
    join_token = ',"~",'
    push_commands_max = 1
    error_description = {
        0: "No Error",
        1: "Over-current condition on output 1",
        2: "Over-current condition on output 2",
        6: "An I/O is already set to this type. Applies to non-General Purpose I/O",
        8: "Tried to SET IO to an incorrect I/O type",
        9: "Tried to write to I/O set as Input or is 'TYPED'",
        10: "Illegal I/O number",
        11: "Incorrect CLOCK type",
        12: "Input 1 not defined as a capture input",
        20: "Tried to set unknown variable or flag. Trying to set an undefined variable of flag. Also could be a typo.",
        21: "Tried to set an incorrect value. Many variables have a range such as the Run Current (RC) which is 1 to 100%. As an example, you cannot set the RC to 110%.",
        22: "VI is set greater than or equal to VM. The Initial Velocity is set equal to, or higher than the Maximum Velocity. VI must be less than VM.",
        23: "VM is set less than or equal to VI. The Maximum Velocity is set equal to, or lower than the Initial Velocity. VM must be greater than VI.",
        24: "Illegal data entered. Data has been entered that the devices does not understand.",
        25: "Variable or flag is read only. Read only flags and variables cannot be set.",
        26: "Variable or flag is not allowed to be incremented or decremented. IC and DC cannot be used on variables or flags such as Baud and Version.",
        27: "Trip not defined. Trying to enable a trip that has not yet been defined.",
        28: "WARNING! Trying to redefine a program label or variable.",
        29: "Trying to redefine a built in command, variable or flag.",
        30: "Unknown label or user variable. Trying to Call or Branch to a Label or Variable that has not yet been defined.",
        31: "Program label or user variable table is full. The table has a maximum capacity of 22 labels and/or user variables.",
        32: "Trying to set a label (LB).",
        33: "Trying to SET an Instruction.",
        34: "Trying to Execute a Variable or Flag",
        35: "Trying to Print Illegal Variable or Flag",
        36: "Illegal Motor Count to Encoder Count Ratio",
        37: "Command, Variable or Flag Not Available in Drive",
        38: "Missing parameter separator",
        39: "Trip on Position and Trip on Relative Distance not allowed together",
        40: "Program not running. If HOLD (H) is entered in Immediate Mode and a program is not running.",
        41: "Stack overflow",
        42: "Illegal program card_address. Tried to Clear, List, Execute, etc. an incorrect Program card_address.",
        43: "Tried to overflow program stack. Calling a Sub-Routine or Trip Routine with no Return.",
        44: "Program locked. User Programs can be Locked with the <LK> command. Once Locked, the program cannot be listed or edited in any way.",
        45: "Trying to Overflow Program Space.",
        46: "Not in Program Mode.",
        47: "Tried to Write to Illegal Flash Address",
        48: "Program Execution stopped by I/O set as Stop.",
        60: "Tried to enter an unknown command",
        61: "Trying to set illegal BAUD rate. The only Baud Rates accepted are those listed on the Properties Page of IMS Terminal.",
        62: "IV already pending or IF Flag already TRUE.",
        63: "Character over-run. Character was received. Processor did not have time to process it and it was over-written by the next character.",
        70: "FLASH Check Sum Fault",
        71: "Internal Temperature Warning",
        72: "Internal Over Temperature disabling drive",
        73: "Tried to SAVE while moving",
        74: "Tried to Initialize Parameters (IP) or Clear Program (CP) while Moving",
        76: "Microstep resolution set to low, must be greater than min sys. speed.",
        77: "VM, VI, or SL too great for selected microstep resolution",
        78: "Aux Power out of range",
        79: "V+ out of range",
        80: "HOME switch not defined. Attempting to do a HOME (H) sequence but the Home Switch has not yet been defined.",
        81: "HOME type not defined. The HOME (HM or HI) Command has been programmed butwith no type or an illegal type. (Types = 1, 2, 3, or 4)",
        82: "Went to both LIMITS and did not find home. The motion encroached both limits but did not trip the Home switch. Indicates a possible bad switch or a bad circuit.",
        83: "Reached plus LIMIT switch. The LIMIT switch in the plus direction was tripped.",
        84: "Reached minus LIMIT switch. The LIMIT switch in the minus direction was tripped.",
        85: "MA or MR isn't allowed during a HOME and a HOME isn't allowed while the devices is in motion.",
        86: "Stall detected. The Stall Flag (ST) has been set to 1.",
        87: "Not allowed to change AS mode while moving",
        88: "Moves not allowed while calibration is in progress.",
        89: "Calibration not allowed while motion is in progress.",
        90: "Motion Variables are too low switching to EE=1",
        91: "Motion stopped by I/O set as Stop.",
        92: "Position Error in Closed loop. motor will attempt to position the shaft within the dead band, After failing 3 attempts Error 92 will be generated. Axis will continue to function normally.",
        93: "MR or MA not allowed while correcting position at end of previous MR or MA.",
        94: "Motion commanded while drive disabled (DE=0)",
        95: "Rotation of direction (RD) attempted while axis is in motion",
        100: "Configuration test done, encoder resolution mismatch",
        101: "Configuration test done, encoder direction incorrect",
        102: "Configuration test done, encoder resolution and direction incorrect",
        103: "Configuration not done, drive not enabled",
        104: "Locked rotor",
        105: "Maximum position count reached",
        106: "Lead limit reached",
        107: "Lag limit reached",
        108: "Lead/lag not zero at the end of a move",
        109: "Calibration failed because drive not enabled.",
        110: "Make up disabled.",
        111: "Factory calibration failed"
    }

    def __init__(self, host, port=503, device_id=None, config=None):
        super().__init__(host, port, device_id, config)

    def _init_attributes(self):
        super(DeviceLexium, self)._init_attributes()
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_TEMPERATURE, readonly=True, default_type=np.float16, decimals=2, unit="degC")
        self.create_attribute(ATTR_CHANNELS, default_value=1, readonly=True)

    def handle_configuration(self):

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.get_error_flag()
        self.get_serial_number(callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DeviceLexium, self).handle_readout(readout_buffer)
        self.handle_command(command, output)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, DeviceLexiumChannel):
                    kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)
                    break

    def handle_command(self, command, output):

        command = command.strip()
        data = output.replace(command, "").strip()
        if command == "PR SN":
            self.set_raw_value(ATTR_SERIAL_NUMBER, data)
        elif command == "PR IT":
            self.set_raw_value(ATTR_TEMPERATURE, float(data.replace(", ", ".")))
        elif command == "PR EF":
            if int(data) == 1:
                self.handle_command_error(command, output)
        elif command == "PR ER":
            error_message = self.error_description.get(int(data), u"Undefined error")
            self.logger.error(error_message)
            self.set_value(ATTR_LAST_ERROR, error_message)
            self.notify((0, ATTR_LAST_ERROR), error_message)
            self.notify((0, ATTR_ERROR_CODE), int(data))
        elif command == "PR P":
            self.notify((0, ATTR_POSITION), int(data))
        elif command == "PR ST":
            self.notify((0, ATTR_STALL_FLAG), bool(int(data)))
        elif command == "PR MV":
            self.notify((0, ATTR_MOVING), bool(int(data)))
        elif command == "PR V":
            self.notify((0, ATTR_VELOCITY), int(data))
        elif command == "PR I3":
            self.notify((0, ATTR_POSITIVE_LIMIT_SWITCH), bool(int(data)))
        elif command == "PR I4":
            self.notify((0, ATTR_NEGATIVE_LIMIT_SWITCH), bool(int(data)))
        elif command == "PR I6":
            self.notify((0, ATTR_INDEX_MARK), bool(int(data)))
        elif command == "PR HC":
            self.notify((0, ATTR_HOLDING_CURRENT), int(data))
        elif command == "PR VM":
            self.notify((0, ATTR_MAXIMUM_VELOCITY), int(data))
        elif command == "PR VI":
            self.notify((0, ATTR_INITIAL_VELOCITY), int(data))
        elif command == "PR RC":
            self.notify((0, ATTR_RUN_CURRENT), int(data))
        elif command == "PR RD":
            self.notify((0, ATTR_ROTATION_DIRECTION), int(data))
        elif command == "PR TS":
            self.notify((0, ATTR_TORQUE_SPEED), int(data))
        elif command == "PR EE":
            self.notify((0, ATTR_ENCODER_ENABLED), int(data))
        elif command == "PR MS":
            self.notify((0, ATTR_MICROSTEP_RESOLUTION), int(data))
        elif command == "PR SF":
            self.notify((0, ATTR_STALL_FACTOR), int(data))
        elif command == "PR IS":
            for line in data.split("\r\n"):
                out = re.search("IS = ([0-9]), ([0-9]), ([0-9])", line)
                if out:
                    if out.group(1) == "3":
                        if int(out.group(2)) == 0:
                            self.notify((0, ATTR_USE_POSITIVE_LIMIT_SWITCH), False)
                        else:
                            self.notify((0, ATTR_USE_POSITIVE_LIMIT_SWITCH), bool(int(out.group(3))))
                    elif out.group(1) == "4":
                        if int(out.group(2)) == 0:
                            self.notify((0, ATTR_USE_POSITIVE_LIMIT_SWITCH), False)
                        else:
                            self.notify((0, ATTR_USE_NEGATIVE_LIMIT_SWITCH), bool(int(out.group(3))))
                        break

    def start_polling(self):
        DeviceSocket.start_polling(self)
        self.poll_command("PR IT\r", 1000)

    def get_serial_number(self, callback=None):
        return self.command(u"PR SN\r", with_token=True, callback=callback)

    def get_device_temperature(self, callback=None):
        return self.command(u"PR IT\r", with_token=True, callback=callback)

    def get_error_flag(self, callback=None):
        return self.command(u"PR EF\r", with_token=True, callback=callback)

    def get_error_code(self, callback=None):
        return self.command(u"PR ER\r", with_token=True, callback=callback)

    def collect_incoming_data(self, data):
        if data.find(b"?") != -1:
            (command, token, callback, returning), timestamp = self.commands_buffer.popleft()
            self.handle_command_error(command, "".join(self.buffer))
            self.device_poller.prepend_command(self, (u"PR ER\r", token, callback, returning))
            self.buffer = []
        else:
            super().collect_incoming_data(data)


class DeviceLexiumChannel(DeviceChannel):
    available_resolutions = [1, 2, 4, 5, 8, 10, 16, 25, 32, 50, 64, 100, 108, 125, 127, 128, 180, 200, 250, 256]
    join_token = ',"~",'

    def __init__(self, device, channel, device_id=None, config=None):
        self.referencing_stop = Event()
        self.referencing_stop.set()
        self.backlash_cleared_flag = True
        super(DeviceLexiumChannel, self).__init__(device, channel, device_id, config)

    def _init_attributes(self):
        super(DeviceLexiumChannel, self)._init_attributes()
        self.create_attribute(ATTR_POSITION, default_type=np.float64, readonly=False, decimals=3,
                              set_function=self._move_absolute, set_value_when_set_function=False)
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_type=np.float64, decimals=3,
                              set_function=self.set_lower_position_limit)
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_type=np.float64, decimals=3,
                              set_function=self.set_upper_position_limit)
        self.create_attribute(ATTR_BACKLASH_DISTANCE, default_value=0, default_type=np.float64, readonly=False,
                              decimals=3)
        self.create_attribute(ATTR_STALL_FLAG, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_MOVING, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_VELOCITY, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_HOLDING_CURRENT, default_type=np.uint8, min_value=0, max_value=100, unit="%",
                              set_function=self.set_holding_current)
        self.create_attribute(ATTR_INITIAL_VELOCITY, default_type=np.uint64, unit="steps/s",
                              set_function=self.set_initial_velocity)
        self.create_attribute(ATTR_MAXIMUM_VELOCITY, default_type=np.uint64, unit="steps/s",
                              set_function=self.set_maximum_velocity)
        self.create_attribute(ATTR_RUN_CURRENT, default_type=np.uint8, max_value=100, unit="%",
                              set_function=self.set_run_current)
        self.create_attribute(ATTR_ROTATION_DIRECTION, default_type=np.uint8, max_value=1,
                              set_function=self.set_rotation_direction)
        self.create_attribute(ATTR_TORQUE_SPEED, default_type=np.uint64, set_function=self.set_torque_speed)
        self.create_attribute(ATTR_ENCODER_ENABLED, default_type=np.uint8, min_value=0, max_value=1,
                              set_function=self.set_encoder_enabled)
        self.create_attribute(ATTR_MICROSTEP_RESOLUTION, default_value="None", default_type=self.available_resolutions,
                              set_function=self.set_microstep_resolution, unit="ustep/step")
        self.create_attribute(ATTR_ENCODER_RESOLUTION, default_value=4000, default_type=np.uint64, unit="count/rev")
        self.create_attribute(ATTR_STEPS_PER_REVOLUTION, default_value=200, default_type=np.uint64, unit="steps")
        self.create_attribute(ATTR_STALL_FACTOR, default_type=np.uint64, min_value=0, max_value=65000,
                              unit="counts", set_function=self.set_stall_factor)
        self.create_attribute(ATTR_POSITIVE_LIMIT_SWITCH, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_NEGATIVE_LIMIT_SWITCH, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_INDEX_MARK, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_USE_POSITIVE_LIMIT_SWITCH, default_type=np.bool,
                              set_function=lambda value: self.set_digital_input(3, 2, value))
        self.create_attribute(ATTR_USE_NEGATIVE_LIMIT_SWITCH, default_type=np.bool,
                              set_function=lambda value: self.set_digital_input(4, 3, value))
        self.create_attribute(ATTR_REFERENCING, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_USE_MARK_INDEX, default_value=True, default_type=np.bool, readonly=False)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.get_encoder_enabled()
        self.get_holding_current()
        self.get_initial_velocity()
        self.get_maximum_velocity()
        self.get_run_current()
        self.get_rotation_direction()
        self.get_torque_speed()
        self.get_microsteps_resolution()
        self.get_stall_factor()
        self.get_digital_inputs(callback=_finish_configuration)

    def start_polling(self):
        super(DeviceLexiumChannel, self).start_polling()
        self.poll_command(u"PR P\r", 200)
        self.poll_command(u"PR MV\r", 200)
        self.poll_command(u"PR ST\r", 500)
        self.poll_command(u"PR V\r", 500)
        self.poll_command(u"PR I3\r", 1000)
        self.poll_command(u"PR I4\r", 1000)

    def stop_polling(self):
        self.remove_poll_command(u"PR P\r", 200)
        self.remove_poll_command(u"PR MV\r", 200)
        self.remove_poll_command(u"PR ST\r", 500)
        self.remove_poll_command(u"PR V\r", 500)
        self.remove_poll_command(u"PR I3\r", 1000)
        self.remove_poll_command(u"PR I4\r", 1000)

    def subject_update(self, key, value, subject):
        super(DeviceLexiumChannel, self).subject_update(key, value, subject)

        if isinstance(key, tuple) and key[0] == self.channel:
            if key[1] == ATTR_MOVING and self.get_value(ATTR_MOVING) != value:
                if value == 1:
                    self.set_status(STATUS_BUSY)
                else:
                    if self.backlash_cleared_flag:
                        self.set_status(STATUS_IDLE)
                    else:
                        self.move_relative(-self.get_value(ATTR_BACKLASH_DISTANCE))
                        self.command(u"PR MV\r")
                        self.backlash_cleared_flag = True
            elif key[1] == ATTR_POSITION:
                if self.get_value(ATTR_ENCODER_ENABLED):
                    value = (float(value) / self.get_value(ATTR_ENCODER_RESOLUTION)) * 360
                else:
                    value = (float(value) / (self.get_value(ATTR_MICROSTEP_RESOLUTION) * self.get_value(
                        ATTR_STEPS_PER_REVOLUTION))) * 360
            elif key[1] == ATTR_ERROR_CODE:
                if value in (83, 84, 86):
                    self.command(u"SL=0\r")
                self.set_status(STATUS_ERROR)
                return
            elif key[1] == ATTR_STALL_FLAG and value is True:
                self.command(u"ST=0\r")

            self.set_raw_value(key[1], value)

    @expose_method()
    def stop(self, callback=None):
        self.logger.info(u"Stop motor channel movement")
        self.referencing_stop.set()
        self.set_value(ATTR_REFERENCING, False)
        self.backlash_cleared_flag = True
        return self.command(u"SL=0\r", callback, with_token=True)

    def _move_absolute(self, position, callback=None):
        self.logger.info(u"Move motor channel to absolute position {}".format(position))
        try:
            position = float(position)
            position_diff = self[ATTR_POSITION][SETPOINT] - self[ATTR_POSITION][VALUE]
            backlash_distance = self.get_value(ATTR_BACKLASH_DISTANCE)

            if self.backlash_cleared_flag and backlash_distance != 0 and np.sign(backlash_distance) == np.sign(
                    position_diff):
                backlash_distance = self[ATTR_POSITION].remove_offset_factor(backlash_distance)
                position += backlash_distance
                self.backlash_cleared_flag = False
            else:
                self.backlash_cleared_flag = True

            if self.get_value(ATTR_ENCODER_ENABLED):
                position = int(round((position / 360.) * self.get_value(ATTR_ENCODER_RESOLUTION)))
            else:
                position = int(round((position / 360.) * (
                        self.get_value(ATTR_MICROSTEP_RESOLUTION) * self.get_value(ATTR_STEPS_PER_REVOLUTION))))
            token = self.command(u"MA {}\r".format(position), callback, with_token=True)
            self.get_moving_state()
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def _move_relative(self, position, callback=None):
        self.logger.info(u"Move motor channel by relative position {}".format(position))
        try:
            position = float(position)
            backlash_distance = self.get_value(ATTR_BACKLASH_DISTANCE)

            if self.backlash_cleared_flag and backlash_distance != 0 and np.sign(backlash_distance) == np.sign(
                    position):
                backlash_distance = self[ATTR_POSITION].remove_offset_factor(backlash_distance)
                position += backlash_distance
                self.backlash_cleared_flag = False
            else:
                self.backlash_cleared_flag = True

            if self.get_value(ATTR_ENCODER_ENABLED):
                position = int(round((position / 360.) * self.get_value(ATTR_ENCODER_RESOLUTION)))
            else:
                position = int(round((position / 360.) * (
                        self.get_value(ATTR_MICROSTEP_RESOLUTION) * self.get_value(ATTR_STEPS_PER_REVOLUTION))))
            token = self.command(u"MR {}\r".format(position), callback, with_token=True)
            self.get_moving_state()
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position, callback=None):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move motor channel to absolute position {}".format(position))
            return self.set_value(ATTR_POSITION, float(position.m), callback)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position, callback=None):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move motor channel by relative position {}".format(position))
            return self.move_absolute(self[ATTR_POSITION][VALUE] + float(position.m), callback)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    def set_lower_position_limit(self, value, callback=None):
        return self.set_attribute([ATTR_POSITION, MIN], value, callback)

    def set_upper_position_limit(self, value, callback=None):
        return self.set_attribute([ATTR_POSITION, MAX], value, callback)

    @expose_method({"position": ATTR_POSITION})
    def reset_position(self, position=0, callback=None):
        try:
            position = device_units(self, ATTR_POSITION, units.Quantity(position))
            self.logger.info(u"Resetting position to {}".format(position))
            position = self[ATTR_POSITION].remove_offset_factor(float(position.m))
            if self.get_value(ATTR_ENCODER_ENABLED):
                position = int(round((position / 360.) * self.get_value(ATTR_ENCODER_RESOLUTION)))
            else:
                position = int(round((position / 360.) * (
                        self.get_value(ATTR_MICROSTEP_RESOLUTION) * self.get_value(ATTR_STEPS_PER_REVOLUTION))))

            return self.command(u"P={}\r".format(position), callback, with_token=True)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"velocity": "steps / second"})
    def slew(self, velocity, callback=None):
        self.logger.info(u"Slew motor channel by velocity {}".format(velocity))
        try:
            return self.command(u"SL={}\r".format(int(velocity)), callback, with_token=True)
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    @expose_method()
    def find_reference_in_positive_direction(self):
        self.stop_polling()
        self.poll_command(u"PR MV\r", 400)

        if self.get_value(ATTR_REFERENCING):
            return False
        Thread(target=self.find_reference, args=(1,)).start()
        return True

    @expose_method()
    def find_reference_in_negative_direction(self):
        self.stop_polling()
        self.poll_command(u"PR MV\r", 400)

        if self.get_value(ATTR_REFERENCING):
            return False
        Thread(target=self.find_reference, args=(-1,)).start()
        return True

    def find_reference(self, direction):
        self.referencing_stop.clear()
        self.set_value(ATTR_REFERENCING, True)

        stage = 0
        maximum_velocity = self.get_value(ATTR_MAXIMUM_VELOCITY)
        limit_switch = ATTR_POSITIVE_LIMIT_SWITCH if direction == 1 else ATTR_NEGATIVE_LIMIT_SWITCH
        if direction == 1:
            self.poll_command(u"PR I3\r", 200)
        else:
            self.poll_command(u"PR I4\r", 200)

        while not self.referencing_stop.wait(0.1):
            if stage == 0:  # Drive with maximum veloctiy to the HW limit
                self.slew(maximum_velocity * direction)
                stage = 1
            elif stage == 1 and self.get_value(limit_switch):  # Check if we are at the limit
                time.sleep(0.3)
                self.slew(maximum_velocity * 0.3 * -direction)
                stage = 2
            elif stage == 2 and not self.get_value(limit_switch):
                time.sleep(0.3)
                self.slew(maximum_velocity * 0.3 * direction)
                stage = 3
            elif stage == 3 and self.get_value(limit_switch):
                if self.get_value(ATTR_USE_MARK_INDEX):
                    self.poll_command(u"PR I6\r", 200)
                    time.sleep(0.3)
                    self.find_index_mark(-direction)
                else:
                    self.slew(maximum_velocity * 0.05 * -direction)
                stage = 4
            elif stage == 4 and self.get_value(ATTR_USE_MARK_INDEX) and self.get_value(ATTR_INDEX_MARK):
                self.remove_poll_command(u"PR I6\r", 200)
                self.stop()
                time.sleep(0.3)
                position = units.Quantity(self.config.get("reference_position", 0))
                self.reset_position(str(position))
            elif stage == 4 and not self.get_value(ATTR_USE_MARK_INDEX) and not self.get_value(limit_switch):
                self.stop()
                time.sleep(0.3)
                position = units.Quantity(self.config.get("reference_position", 0))
                self.reset_position(str(position))

        self.set_value(ATTR_REFERENCING, False)
        self.remove_poll_command(u"PR MV\r", 400)
        self.remove_poll_command(u"PR I3\r", 200)
        self.remove_poll_command(u"PR I4\r", 200)
        self.remove_poll_command(u"PR I6\r", 200)
        self.start_polling()

    def find_index_mark(self, direction):
        if direction == 1:
            direction = 2
        else:
            direction = 4
        self.command(u"HI={}\r".format(direction))

    def set_holding_current(self, value, callback=None):
        self.logger.info(u"Set motor channel holding current to {}%".format(value))
        try:
            token = self.command(u"HC {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_initial_velocity(self, value, callback=None):
        self.logger.info(u"Set motor channel initial velocity to {}".format(value))
        try:
            token = self.command(u"VI {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_maximum_velocity(self, value, callback=None):
        self.logger.info(u"Set motor channel maximum velocity to {}".format(value))
        try:
            token = self.command(u"VM {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_run_current(self, value, callback=None):
        self.logger.info(u"Set motor channel run current to {}%".format(value))
        try:
            token = self.command(u"RC {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_rotation_direction(self, value, callback=None):
        self.logger.info(u"Set motor channel rotation direction to {}".format(value))
        try:
            token = self.command(u"RD {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_torque_speed(self, value, callback=None):
        self.logger.info(u"Set motor channel torque speed to {}".format(value))
        try:
            token = self.command(u"TS {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_stall_factor(self, value, callback=None):
        self.logger.info(u"Set motor channel stall factor to {}".format(value))
        try:
            token = self.command(u"SF {}\r".format(int(value)), callback, True)
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_encoder_enabled(self, value, callback=None):
        self.logger.info(u"Set motor channel encoder enabled to {}".format(value))
        try:
            token = self.command(u"EE {}\r".format(int(value)), callback, True)
            self.get_encoder_enabled()
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_microstep_resolution(self, value, callback=None):
        self.logger.info(u"Set motor channel microstep resolution to {} microsteps / step".format(value))
        try:
            token = self.command(u"MS {}\r".format(int(value)), callback, True)
            self.get_maximum_velocity()
            return token
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def set_digital_input(self, input, type, value):
        self.logger.info(u"Set motor channel digital input {} to {}".format(input, value))
        try:
            self.command(u"IS={},{},{}\r".format(int(input), int(type), int(value)))
            self.get_digital_inputs()
            self.get_on_positive_switch()
            self.get_on_negative_switch()
        except ValueError as e:
            self.logger.exception(e)
            raise DeviceError(e)

    def get_moving_state(self, callback=None):
        return self.command(u"PR MV\r", with_token=True, callback=callback)

    def get_holding_current(self, callback=None):
        return self.command(u"PR HC\r", with_token=True, callback=callback)

    def get_initial_velocity(self, callback=None):
        return self.command(u"PR VI\r", with_token=True, callback=callback)

    def get_maximum_velocity(self, callback=None):
        return self.command(u"PR VM\r", with_token=True, callback=callback)

    def get_run_current(self, callback=None):
        return self.command(u"PR RC\r", with_token=True, callback=callback)

    def get_rotation_direction(self, callback=None):
        return self.command(u"PR RD\r", with_token=True, callback=callback)

    def get_torque_speed(self, callback=None):
        return self.command(u"PR TS\r", with_token=True, callback=callback)

    def get_encoder_enabled(self, callback=None):
        return self.command(u"PR EE\r", with_token=True, callback=callback)

    def get_microsteps_resolution(self, callback=None):
        return self.command(u"PR MS\r", with_token=True, callback=callback)

    def get_stall_factor(self, callback=None):
        return self.command(u"PR SF\r", with_token=True, callback=callback)

    def get_on_index_mark(self, callback=None):
        return self.command(u"PR I6\r", with_token=True, callback=callback)

    def get_on_positive_switch(self, callback=None):
        return self.command(u"PR I3\r", with_token=True, callback=callback)

    def get_on_negative_switch(self, callback=None):
        return self.command(u"PR I4\r", with_token=True, callback=callback)

    def get_digital_inputs(self, callback=None):
        return self.command(u"PR IS\r", with_token=True, callback=callback)

    def get_position(self, callback=None):
        return self.command(u"PR P\r", with_token=True, callback=callback)

    def get_velocity(self, callback=None):
        return self.command(u"PR V\r", with_token=True, callback=callback)



import argparse
import os
import signal
from multiprocessing import Event

import oyaml as yaml

import kamzik3
from kamzik3.devices.deviceSession import server_process
from kamzik3.snippets.snippetsWidgets import init_qt_app

stopped = Event()

if __name__ == "__main__":
    # Create sub server processes
    server_A = server_process("./conf_multi_server_A.yml", stopped)
    server_B = server_process("./conf_multi_server_B.yml", stopped)
    # Start both servers
    server_A.start()
    server_B.start()

    # Parse arguments from commandline
    parser = argparse.ArgumentParser()
    parser.add_argument("--conf", help="Path to configuration file in yaml format",
                        default="./conf_proxy_server.yml")
    parser.add_argument("--chdir", help="Path to active directory",
                        default="./")
    args = parser.parse_args()

    # Set active directory
    os.chdir(args.chdir)
    # Create PyQT5 application
    app = init_qt_app(enable_hd_scaling=False)

    # Load configuration file
    with open(args.conf, "r") as configFile:
        config = yaml.load(configFile, Loader=yaml.Loader)
    # Start control loop for devices specified in configuration file
    kamzik3.session.set_config(args.conf)
    kamzik3.session.start_control_loops()

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    if "session_window" in config:
        config["session_window"].show()

    # Start PyQT5 app loop
    app.exec_()
    # Close session
    kamzik3.session.stop()
    # Set stop event
    stopped.set()
    # Wait for Server A and B to close
    server_A.join()
    server_B.join()




import pyqtgraph as pg
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QMdiSubWindow

from kamzik3.snippets.snippetsUnits import get_scaled_time_duration

pg.setConfigOption(u'background', u'w')
pg.setConfigOption(u'foreground', u"k")


class TimeAxisItem(pg.AxisItem):
    '''
    This class converts unix timestamps into HH:mm format for X axis.
    '''

    x_values = None

    def __init__(self, *args, **kwargs):
        pg.AxisItem.__init__(self, *args, **kwargs)

    def tickStrings(self, values, scale, spacing):
        '''
        Convert all timestamps into HH:mm Strings.
        @param values: List of timestamps
        @param scale: float
        @param spacing: float
        '''
        if self.x_values is None or len(self.x_values) == 0:
            return []
        values = values - self.x_values[0]
        return [get_scaled_time_duration(value, short=True) for value in values]


class LogPlotView(pg.PlotWidget):
    sig_repaint = pyqtSignal()

    def __init__(self, parent=None, background='default', **kwargs):
        self.time_Axis = TimeAxisItem(orientation='bottom')
        super().__init__(parent, background, axisItems={'bottom': self.time_Axis}, **kwargs)
        try:
            self.base_unit = kwargs.get("unit", "")
        except TypeError:
            self.base_unit = ""

        pen1 = pg.mkPen(color=kwargs.get("color", (0, 0, 255)))
        self.value_curve = pg.PlotDataItem([], [], pen=pen1)
        self.setLabel(u"left", kwargs.get("left_label", ""), units=self.base_unit)
        self.setLabel(u"bottom", kwargs.get("bottom_label", None))
        self.setTitle(kwargs.get("top_label", None))
        self.showGrid(x=True, y=True, alpha=0.1)
        self.getAxis("left").enableAutoSIPrefix(False)
        self.addItem(self.value_curve)
        self.sig_repaint.connect(self.slot_set_data)

    @pyqtSlot()
    def slot_set_data(self, x, y):
        self.time_Axis.x_values = x
        self.value_curve.setData(x, y)

    def close(self):
        self.reset()
        super().close()

    def reset(self):
        self.x_buffer.clear()
        self.y_buffer.clear()
        self.value_curve.setData([], [])


class LogPlotViewMaster(LogPlotView):

    def __init__(self, parent=None, background='default', **kwargs):
        super().__init__(parent, background, **kwargs)

        self.lookingGlass = pg.LinearRegionItem([0, 0])
        self.lookingGlass.setZValue(-10)
        self.addItem(self.lookingGlass)

    @pyqtSlot()
    def slot_set_data(self, x, y):
        self.time_Axis.x_values = x
        self.value_curve.setData(x, y)
        if len(x) > 0:
            zoom_from, zoom_to = int(len(x) * 0.4), int(len(x) * 0.6)
            self.lookingGlass.setRegion([x[zoom_from], x[zoom_to]])

    def reset_zoom(self):
        zoom_from, zoom_to = 0, len(self.value_curve.xData) - 1
        self.lookingGlass.setRegion([self.value_curve.xData[zoom_from], self.value_curve.xData[zoom_to]])


class QMdiSubWindowPlot(QMdiSubWindow):

    def __init__(self, id, parent=None):
        self.id = id
        QMdiSubWindow.__init__(self, parent)



import logging
import time
from collections import deque
from copy import copy
from math import inf
from threading import Thread

import numpy as np
import serial
from kamzik3.snippets.snippetsDecorators import expose_method
from pint import UndefinedUnitError, DimensionalityError

import kamzik3
from kamzik3 import WriteException, CommandFormatException, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute, SetFunction
from kamzik3.devices.subject import Subject
from kamzik3.snippets.snippetDataAccess import get_from_dict, fullname, is_equal
from kamzik3.snippets.snippetsControlLoops import control_device_poller, control_device_connection_poller
from kamzik3.snippets.snippetsGenerators import device_id_generator, token_generator
from kamzik3.snippets.snippetsUnits import device_units
from kamzik3.snippets.snippetsYaml import YamlSerializable


class Device(Subject, YamlSerializable):
    """
    Connection is done in sequence:
        : connect()
        : handle_connect_event()
            : handle_connect()
            : handle_configuration_event()
                : handle_configuration()

    Connection error handling:
        : if connect() => error state
           : disconnect() if critical error
           : close(), reconnect() if not critical error
        : if connect() => connection timeout
            : handle_connection_error(), close()
        : if Device.connected and response timeout()
            : handle_response_error(), close_connection()

    Disconnection is done in sequence:
        : disconnect()
            : close() if not Device.connected
            : close_connection() if connected
                : close()
    """
    id_generator = device_id_generator()
    connection_timeout = 4000
    response_timeout = 4000
    session = None
    max_command_retry = 0
    command_retry_count = 0
    push_commands_max = inf
    push_buffer_size = 2 ** 16
    command_encoding = "utf8"

    def __init__(self, device_id=None, config=None):
        self.connecting_time = 0
        self.response_timestamp = 0
        self.request_timestamp = 0
        self.connected = False
        self.connecting = False
        self.closing = False
        self.closed = False
        self.connection_error = False
        self.response_error = False
        self.latency_buffer = []
        self.init_time = time.time()
        self.commands_buffer = deque()
        Subject.__init__(self)
        self.device_id = device_id
        if device_id is None:
            self.device_id = next(Device.id_generator)
        self.config = config
        if self.config is None:
            self.config = {}
        self.token_generator = token_generator()
        self.device_poller = control_device_poller
        self.device_connection_poller = control_device_connection_poller
        # Get absolute path to the Device class
        self.qualified_name = fullname(self)
        if not hasattr(self, "logger"):
            self.logger = logging.getLogger("Device.{}".format(self.device_id))
        if "push_commands_max" in self.config:
            self.push_commands_max = self.config["push_commands_max"]
        # Size of command buffer used in push
        if "push_buffer_size" in self.config:
            self.push_buffer_size = self.config.get("push_buffer_size")
        self.macro_steps = {MACRO_SET_ATTRIBUTE_STEP: ["*"], MACRO_EXECUTE_METHOD_STEP: ["*"]}
        self.exposed_methods = []
        self._expose_methods_to_clients()
        # Check if attributes are already defined
        if not hasattr(self, "attributes"):
            self.attributes = {}
            self.attributes_sharing_map = {}
            self.attribute_attach_map = {}
            self._init_attributes()
        # Check if device id is already registered in session and remove it if it's not our current Device
        if self.device_id in kamzik3.session.devices and self != kamzik3.session.devices[self.device_id]:
            kamzik3.session.devices[self.device_id].close()
            kamzik3.session.devices[self.device_id] = None
            del kamzik3.session.devices[self.device_id]
        # Check if device is not registered in session
        if kamzik3.session is not None and self.device_id not in kamzik3.session.devices:
            self.set_session(kamzik3.session)

    def __getitem__(self, item):
        return self.attributes[item]

    def __setitem__(self, key, value):
        self.attributes[key] = value

    def _init_attributes(self):
        """
        Initiate Device's attributes.
        To call this method is mandatory.
        Overload this method for any other derived Device.
        """

        def set_status(key, value):
            if key == VALUE:
                self.notify(ATTR_STATUS, value)

        self.create_attribute(ATTR_ID, default_value=self.device_id, readonly=True, description="Unique device ID")
        self.create_attribute(ATTR_STATUS, default_value=STATUS_DISCONNECTED, readonly=True,
                              description="Current device status").attach_callback(set_status)
        self.create_attribute(ATTR_DESCRIPTION, readonly=True, description="Device description")
        self.create_attribute(ATTR_ENABLED, default_value=True, default_type=np.bool,
                              description="Allow any external changes to device")
        self.create_attribute(ATTR_LATENCY, default_value=0, default_type=np.uint16,
                              description="Latency between sending command and receiving an answer",
                              min_value=0, max_value=9999, unit="ms", readonly=True)
        self.create_attribute(ATTR_BUFFERED_COMMANDS, default_value=0, default_type=np.uint32,
                              description="Amount of commands waiting to be executed",
                              min_value=0, readonly=True)
        self.create_attribute(ATTR_HANGING_COMMANDS, default_value=0, default_type=np.uint32,
                              description="Amount of commands waiting to be answered from device",
                              min_value=0, readonly=True)
        self.create_attribute(ATTR_LAST_ERROR, readonly=True, description="Last device error exception")

    def create_attribute(self, name, group=None, **kwargs):
        """
        Use this function to create new attribute.
        :param name: str
        :param group: str
        :param kwargs: dict
        :return: Attribute
        """
        attribute_path = name
        if group is not None:
            attribute_path = "{}.{}".format(group, attribute_path)
        attribute_id = "{}.{}.{}".format(self.device_id, TOKEN_ATTRIBUTE, attribute_path)
        attribute = Attribute(attribute_id, **kwargs)

        if group is None:
            self.attributes[name] = attribute
        else:
            if group not in self.attributes:
                self.attributes[group] = {}
            self.attributes[group][name] = attribute
        if kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute.attribute_id, TOKEN_ATTRIBUTE_REPLACE)
            kamzik3.session.publisher.push_message(header, attribute)
        return attribute

    def add_attribute(self, name, attribute, group=None):
        """
        Use this function to add attribute to Device.
        Set group parameter if You want attribute be in specific group.
        :param name: str
        :param attribute: Attribute
        :param group: str
        :return: Attribute
        """
        attribute_list = [name]
        if group is None:
            self.attributes[name] = attribute
        else:
            attribute_list.insert(0, group)
            if group not in self.attributes:
                self.attributes[group] = {}
            self.attributes[group][name] = attribute

        attribute_exists = attribute.attribute_id in self.attributes_sharing_map
        if attribute_exists:
            self.attributes_sharing_map[attribute.attribute_id].append(attribute_list)
        else:
            self.attributes_sharing_map[attribute.attribute_id] = [attribute_list]
        if attribute_exists and kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute.attribute_id, TOKEN_ATTRIBUTE_REPLACE)
            kamzik3.session.publisher.push_message(header, attribute)
        if attribute.attribute_id is None:
            attribute_path = name
            if group is not None:
                attribute_path = "{}.{}".format(group, attribute_path)
            attribute.attribute_id = "{}.{}.{}".format(self.device_id, TOKEN_ATTRIBUTE, attribute_path)
        return attribute

    def delete_attribute(self, name, group=None):
        """
        Delete specific attribute from group
        :param name: str
        :param group: str
        :return: None
        """
        attribute_path = name
        if group is not None:
            attribute_path = "{}.{}".format(group, attribute_path)
        attribute_id = "{}.{}.{}".format(self.device_id, TOKEN_ATTRIBUTE, attribute_path)
        if group is not None:
            del self.attributes[group][name]
        else:
            del self.attributes[name]
        if kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute_id, TOKEN_ATTRIBUTE_DELETE)
            kamzik3.session.publisher.push_message(header, attribute_path)

    def delete_attribute_group(self, group):
        """
        Delete all attributes including group
        :param group: str
        :return: None
        """
        attribute_path = group
        attribute_id = "{}.{}.{}".format(self.device_id, TOKEN_ATTRIBUTE, attribute_path)
        del self.attributes[group]
        if kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute_id, TOKEN_ATTRIBUTE_GROUP_DELETE)
            kamzik3.session.publisher.push_message(header, attribute_path)

    def share_group(self, source_device, source_group, target_group, attribute_name_mask=None):
        """
        Share all attributes in group filtered by attribute_name_mask from source_device.
        :param source_device: Device
        :param source_group: str
        :param target_group: str
        :param attribute_name_mask: dict
        :return: None
        """
        if attribute_name_mask is None:
            attribute_name_mask = {}
        if source_group is not None:
            from_group = source_device.attributes[source_group]
        else:
            from_group = source_device.attributes

        for attribute_name, attribute in from_group.items():
            self.add_attribute(attribute_name_mask.get(attribute_name, attribute_name), attribute, target_group)

    def share_exposed_method(self, source_device, source_method, shared_method_name=None,
                             shared_method_attributes=None):
        """
        Share exposed method with source_device.
        This is handy method when You share attributes of other device and You want to also expose it's methods.
        Each new method will go under new name source_device.id_source_method.name.
        If You define shared_method_attributes they will overwrite default ones.
        :param source_device: Device
        :param source_method: Name of exposed function
        :param shared_method_attributes: Dictionary of method attributes
        :return:
        """
        for method_name, method_parameters in source_device.exposed_methods:
            if method_name == source_method:
                if shared_method_name is None:
                    shared_method_name = "{}_{}".format(source_device.device_id, method_name)
                if shared_method_attributes is None:
                    shared_method_attributes = method_parameters
                method = copy(getattr(source_device, method_name))
                # Check if method is meant to be exposed and remove exposed_parameters to prevent double exposing
                if hasattr(method, "exposed_parameters"):
                    del method.__dict__["exposed_parameters"]
                setattr(self, shared_method_name, method)
                self.exposed_methods.append((shared_method_name, shared_method_attributes))

    def set_session(self, session):
        """
        Set session for device.
        :param session: Session
        :return: None
        """
        self.session = session
        session.add_device(self)

    def attach_attribute_callback(self, attribute, callback, max_update_rate=None, key_filter=None):
        """
        Attach callback for every change of attribute.
        If max_update_rate is specified that callback will be called only after
        specified timeout.
        :param attribute: [ATTRIBUTE]
        :param callback: callable
        :param max_update_rate: millisecond(s)
        :return: None
        """
        assert callable(callback)
        try:
            attribute_object = self.get_attribute(attribute)
            attribute_object.attach_callback(callback, max_update_rate, key_filter)

            try:
                self.attribute_attach_map[attribute_object.attribute_id].append((attribute, callback))
            except KeyError:
                self.attribute_attach_map.update({attribute_object.attribute_id: [(attribute, callback)]})

            if key_filter is not None:
                callback(attribute_object[key_filter])
            else:
                callback(VALUE, attribute_object[VALUE])
        except (KeyError, AttributeError):
            self.logger.exception(u"Attribute {} is not defined".format(attribute))

    def detach_attribute_callback(self, attribute, callback):
        """
        Detach attached callback from attribute.
        :param attribute: [ATTRIBUTE]
        :param callback: callable
        :return: None
        """
        assert callable(callback)
        try:
            attribute_object = self.get_attribute(attribute)
            attribute_object.detach_callback(callback)

            try:
                if (attribute, callback) in self.attribute_attach_map[attribute_object.attribute_id]:
                    self.attribute_attach_map[attribute_object.attribute_id].remove((attribute, callback))
            except KeyError:
                pass  # Callback does not exists

        except (KeyError, AttributeError):
            if self.logger is not None:
                self.logger.exception(u"Attribute {} is not defined".format(attribute))

    def get_attribute(self, attribute):
        """
        Get Attribute value
        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: None
        """
        try:
            return self._get(Attribute.list_attribute(attribute))
        except KeyError:
            return None

    def get_value(self, attribute, key=VALUE):
        """
        Get Attribute[key] value
        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: None
        """
        try:
            return self.get_attribute(attribute)[key]
        except (KeyError, TypeError):
            return None

    def get_raw_value(self, attribute, key=VALUE):
        """
        Get attribute[key] value without offset or factor.
        Use returned value to feed into real hardware.
        :param attribute: list, dict, str
        :return: None
        """
        try:
            attribute_object = self.get_attribute(attribute)
            return attribute_object.remove_offset_factor(attribute_object[key])
        except (KeyError, TypeError):
            return None

    def _get(self, attribute):
        """
        This gets Device attribute.
        Use this function when You want to get attribute by tuple or list key.
        Example: Device.get((ATTR_STATUS, VALUE))
        Otherwise use faster method Device[ATTR_STATUS][VALUE]
        :param attribute: list
        :return: mixed
        """
        try:
            return get_from_dict(self.attributes, attribute)
        except KeyError:
            return None

    def set_attribute(self, attribute, value, callback=None):
        """
        This sets Device attribute.
        Re-implement it for client or any special use.
        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: None
        """
        list_attribute = Attribute.list_attribute(attribute)
        if list_attribute[-1] == VALUE:
            return self.set_value(list_attribute[:-1], value, callback)
        else:
            return self._set(list_attribute, value, callback)

    def set_value(self, attribute, value, callback=None, key=VALUE):
        """
        Set attribute value.
        :param attribute: list, dict, str
        :param value: mixed
        :return: None
        """
        attribute_list = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(attribute_list)
        with SetFunction(attribute_object, callback):
            return self._set(attribute_list + [key], value, callback)

    def set_raw_value(self, attribute, value, callback=None, key=VALUE):
        """
        Set attribute value obtained from device.
        Apply offset and factor before value is set.
        :param attribute: list, dict, str
        :param value: mixed
        :return: None
        """
        list_attribute = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(list_attribute)
        value = attribute_object.apply_offset_factor(value)
        self._set(list_attribute + [key], value, callback)

    def _set(self, attribute_list, value, callback=None):
        """
        This sets Device attribute.
        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        Attribute value is pushed into server if Device is connected listed on one.
        To reduce amount of pushed attributes we check if value is different from previous one.
        :param attribute: list
        :param value: mixed
        :return: None
        """
        attribute_object = get_from_dict(self.attributes, attribute_list[:-1])
        if not attribute_object[ALLOW_STREAM]:
            propagate_change = False
        elif attribute_list[-1] == VALUE and attribute_object.with_set_function:
            propagate_change = attribute_object.set_value_when_set_function
        else:
            propagate_change = True

        current_value = attribute_object[attribute_list[-1]]
        attribute_object[attribute_list[-1]] = value

        set_token = attribute_object.read_and_reset_token()
        if not set_token and callable(callback):
            callback(attribute_list, value)

        if propagate_change:
            value = attribute_object[attribute_list[-1]]
            if not is_equal(current_value, value) and kamzik3.session.publisher is not None:
                if self.session and attribute_object[SAVE_CHANGE]:
                    self.session.update_config(self.device_id, attribute_list, value)
                header = "{}.{}".format(attribute_object.attribute_id, attribute_list[-1])
                kamzik3.session.publisher.push_message(header, value, min_timeout=attribute_object.min_stream_timeout)

        return set_token

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        try:
            self.connecting = True
            self.connected = False
            self.device_connection_poller.add_connecting_device(self)
            self.handle_connect_event()
        except DeviceError:
            self.logger.exception(u"Connection exception")
            return

    def handle_readout_callback(self, callback, attribute, output):
        """

        :param callback: callable
        :param attribute: [ATTRIBUTE]
        :param output: mixed
        :return: None
        """
        if callback is None:
            return
        else:
            assert callable(callback)
            callback(attribute, output)

    def handle_connect_event(self):
        """
        Handle connect event.
        This method is a wrapper in which connection is handled.
        Set everything important in handle_connect().
        """
        try:
            self.handle_connect()
            self.connected = True
            self.connecting = False
            self.handle_configuration_event()
        except DeviceError:
            """
            Here we encountered connection error.
            Log error.
            """
            self.logger.exception(u"Error during connection")

    def handle_connect(self):
        """
        Load configuration in this method.
        """
        self.set_status(STATUS_CONNECTED)
        self.logger.info(u"Device connection took {} sec.".format(time.time() - self.init_time))
        self.request_timestamp = self.response_timestamp = time.time()

    def handle_configuration_event(self):
        """
        Event from connected device to handle configuration.
        It servers as a wrapper around configuration routine.
        :return:
        """
        try:
            self.set_status(STATUS_CONFIGURING)
            self.handle_configuration()
        except DeviceError:
            self.logger.exception(u"Error during configuration")

    def handle_configuration(self):
        """
        Method is called whenever device is connected and ready to be configured
        :return: None
        """
        raise NotImplementedError(u"Must be implemented in subclass")

    def handle_response_error(self, message=None):
        """
        Method to handle response error of device
        :param message: cause of response error
        :return:
        """
        self.logger.error(message)
        self.response_error = True
        self.close_connection()

    def handle_connection_error(self, message=None):
        """
        Method to handle state of connection error
        :param message: cause of connection error
        :return: None
        """
        self.logger.error(message)
        self.connection_error = True
        self.close()

    def handle_command_error(self, readout_command, readout_output):
        """
        Method to handle state of command error
        :param readout_command: original command
        :param readout_output: error response from device
        :return: None
        """
        self.set_value(ATTR_LAST_ERROR, str(readout_output))
        self.logger.error(
            u"Command error\nCommand: {!r}\nOutput: {!r}\nCommand buffer: {!r}".format(readout_command, readout_output,
                                                                                       self.commands_buffer))

    def handle_observer_attached(self, observer):
        """
        Callback whenever observer is detached from the Subject
        :param observer: Observer
        :return: None
        """
        observer.subject_update(ATTR_STATUS, self.get_value(ATTR_STATUS), self)

    def handle_observer_detached(self, observer):
        """
        Callback whenever observer is attached to the Subject
        :param observer: Observer
        :return:
        """
        pass

    def handle_readout(self, readout_buffer):
        """
        We have data in readoutBuffer.
        Return tuple attribute, joined readoutBuffer
        :param readout_buffer:
        :return: str, str
        """
        self.response_timestamp = time.time()

        try:
            (attribute, token, callback, returning), command_init_timestamp = self.commands_buffer.popleft()
            latency = (time.time() - command_init_timestamp) * 1000
            self.latency_buffer.append(latency)
            if len(self.latency_buffer) == 20:
                self.set_attribute((ATTR_LATENCY, VALUE), sum(self.latency_buffer) / 20.)
                self.set_value(ATTR_HANGING_COMMANDS, len(self.commands_buffer))
                self.latency_buffer = []
        except IndexError:
            self.handle_response_error(
                u"Trying to pop from empty command buffer. Content of readout buffer is: {}".format(readout_buffer))
            return RESPONSE_ERROR, "", None, False

        if self.command_encoding is None:
            return attribute, readout_buffer, callback, token
        else:
            return attribute, "".join(readout_buffer), callback, token

    def disconnect(self):
        """
        Call this function to cleanly close connection.
        After disconnect Device won't reconnect.
        :return:
        """
        if self.closing:
            return False
        elif self.connected:
            self.close_connection()
        elif self.connecting:
            self.close()

    def close(self):
        """
        Handle all devices closing stuff here.
        Close socket, close port, etc...
        """
        if self.closing:
            return False
        else:
            self.closing = True
            self.connected = False
            self.set_status(STATUS_DISCONNECTED)
            self.closing = False
            self.connected = False

    def reconnect(self):
        """
        Reconnect devices.
        If Device was reconnect return True, False otherwise.
        :return: bool
        """
        mapping = self.yaml_mapping()
        self.__init__(**mapping)
        return True

    def set_status(self, status):
        """
        Set attributes status value.
        :param status: str
        :return: None
        """
        self.set_value(ATTR_STATUS, status)

    def is_status(self, status):
        """
        Check status value against current device status
        :param status: STATUS
        :return: bool
        """
        return self.get_value(ATTR_STATUS) == status

    def in_statuses(self, statuses):
        """
        Check list of statuses against current device status
        :param statuses: list, tuple
        :return: bool
        """
        return self.get_value(ATTR_STATUS) in statuses

    def stop_polling(self):
        """
        Remove Device from device_poller loop.
        """
        self.device_poller.stop_polling(self)

    def start_polling(self):
        """
        Add Device attributes to be polled in devicePoller loop.
            self.devicePoller.add(devices, command, time in milliseconds)
            self.devicePoller.add(self, 'cmd0', 400)
        """
        self.response_timestamp = self.request_timestamp = time.time()

    def close_connection(self):
        """
        Call this function when closing connected Device.
        Stop polling and continue with close() method.
        """
        self.stop_polling()
        self.close()

    def push(self, data):
        """
        Send command to Device.
        This method will send data directly to the Device connection line.
        Use command() for sending commands from clients, etc.
        :param data: str
        """
        pass

    def command(self, command, callback=None, with_token=False, returning=True):
        """
        Send command to devices.
        Use only this function to send command to Device.
        Commands are stored into buffer and are flushed every pollDevice tick.
        :param command:
        :param callback:
        :param with_token:
        :param returning: bool
        :return:
        """
        if not self.valid_command_format(command):
            raise CommandFormatException(u"Command '{}' form is invalid".format(command))

        if self.connected or self.connecting:
            token = 0
            if with_token is True:
                token = next(self.token_generator)
            elif with_token >= 1:
                token = with_token
            self.device_poller.prepare_command(self, (command, token, callback, returning))
            return token

    def remove(self):
        """
        This method aims to remove device from runtime.
        :return:
        """
        self.disconnect()

        if self.logger is not None:
            for handler in self.logger.handlers[:]:
                handler.close()

        self.logger = None
        if self.session is not None:
            self.session.remove_device(self)

        for cb_tuples in self.attribute_attach_map.values():
            for cb_tuple in cb_tuples:
                self.detach_attribute_callback(*cb_tuple)

        self.attribute_attach_map = None
        self.attributes = None
        self.exposed_methods = None
        self.token_generator = None
        self.connecting_time = None
        self.response_timestamp = None
        self.request_timestamp = None
        self.connected = None
        self.connecting = None
        self.closing = None
        self.closed = None
        self.connection_error = None
        self.response_error = None
        self.commands_buffer = None

    def valid_command_format(self, command):
        """
        Check if command format is valid.
        Return False or optionally raise CommandFormatException.
        Re-implement this method for any other device.
        :param command:
        :return:
        """
        if command is not None:
            return True
        else:
            return False

    def send_command(self, commands):
        """
        Send command using defined interface (TCP/IP, RS232, ...).
        Unicode command is always .encode() into ASCII before it's send to the interface.
        :param commands: list of command to be send
        :return: list of NOT sent commands
        """
        if not self.accepting_commands():
            return commands
        try:
            mark = time.time()
            if self.connected:
                command_data = b"" if self.command_encoding is None else ""
                commands_to_push_counter = 0
                while commands:
                    command = commands.pop(0)
                    commands_to_push_counter += 1
                    command_data += command[0]
                    if len(command_data) > self.push_buffer_size:
                        commands.insert(0, command)
                        command_data = command_data[:-len(command[0])]
                        break
                    if command[3]:  # If command is returning
                        self.commands_buffer.append((command, mark))
                    else:  # Command is not returning, simulate immediate execution
                        self.response_timestamp = mark
                    if commands_to_push_counter >= self.push_commands_max:
                        break
                self.push(command_data if self.command_encoding is None else command_data.encode(self.command_encoding))
            else:
                self.handle_response_error(u"Device is not connected")
        except IndexError:
            self.handle_connection_error(u"Device {} buffer error".format(self.device_id))
        except (WriteException, serial.SerialException):
            self.handle_response_error(u"Device {} writing error".format(self.device_id))
        finally:
            return commands

    def accepting_commands(self):
        """
        Check if device is ready to send another commands.
        :return: bool
        """
        self.request_timestamp = time.time()
        if not self.connected or len(self.commands_buffer) > 0:
            return False
        else:
            return True

    def poll_command(self, command, interval):
        """
        Add command to device poller to be polled in desired interval.
        :param command: unicode string
        :param interval: interval in ms
        :return:
        """
        self.device_poller.add(self, command, interval)

    def remove_poll_command(self, command, interval):
        """
        Remove command from device poller.
        :param command:
        :param interval:
        :return:
        """
        self.device_poller.remove(self, command, interval)

    def _config_attributes(self):
        """
        Set attributes to desired values found in config.
        :return: None
        """
        if self.config:
            for attribute, value in self.config.get("attributes", {}).items():
                unit = self.get_attribute(attribute[:-1])[UNIT]
                if unit in (None, "") or attribute[-1] == UNIT:
                    self.set_attribute(attribute, value)
                else:
                    if isinstance(value, list):
                        set_value = []
                        for v in value:
                            try:
                                set_value.append(device_units(self, list(attribute[:-1]), v).m)
                            except (UndefinedUnitError, DimensionalityError):
                                set_value.append(v)
                    else:
                        try:
                            set_value = device_units(self, list(attribute[:-1]), value).m
                        except (UndefinedUnitError, DimensionalityError):
                            set_value = value
                    self.set_attribute(attribute, set_value)

    def _config_commands(self):
        """
        Execute required methods found in commands section of config
        :return: None
        """
        if self.config:
            for command in self.config.get("commands", []):
                self.command(command)

    def _expose_methods_to_clients(self):
        """
        This method is called to expose all marked methods to the clients.
        :return: None
        """
        for method in dir(self):
            method = getattr(self, method)
            if callable(method) and hasattr(method,
                                            "exposed_parameters") and method.__name__ not in self.exposed_methods:
                self.exposed_methods.append((method.__name__, method.exposed_parameters))

    def wait_for_status(self, statuses, retry_timeout=1000, throw_exception=True, callback=None):
        """
        Wait until device change it's status in one defined in statuses.
        If callback is set, then method is executed in non blocking thread.
        :param statuses: List o one or multiple statuses to check against
        :param retry_timeout: Maximum time in ms to wait for status
        :param callback: Function to call after check in non blocking execution
        :return: True if device is in one of statuses
        """

        def _wait(_callback=None):
            success = self.in_statuses(statuses)
            retry_counter = 0
            while not success:
                if not self.connected:
                    break
                time.sleep(0.05)
                retry_counter += 50
                success = self.in_statuses(statuses)
                if retry_counter >= retry_timeout:
                    break
            if _callback is not None:
                _callback(success)
            elif throw_exception:
                if not success:
                    raise DeviceError(
                        "Status was not set to {} within timeout of {} ms".format(statuses, retry_timeout))
                else:
                    return True
            else:
                return success

        if callback is None:
            return _wait()
        else:
            thread = Thread(target=_wait, args=[callback])
            thread.start()
            return thread

    def wait_until_value_set(self, attribute, value, retry_timeout=1000, throw_exception=True, callback=None):
        """
        This function triggers callback or block until attribute is set to requested value.
        There is retry_timeout of 1s until DeviceError is raised.
        :param attribute:
        :param value:
        :param retry_timeout:
        :param callback:
        :return:
        """

        def _wait(_callback=None):
            attribute_object = self.get_attribute(attribute)
            success = attribute_object[VALUE] == value
            retry_counter = 0
            while not success:
                if not self.connected:
                    break
                time.sleep(0.05)
                retry_counter += 50
                success = attribute_object[VALUE] == value
                if retry_counter >= retry_timeout:
                    break
            if _callback is not None:
                _callback(success)
            elif throw_exception:
                if not success:
                    raise DeviceError(
                        "Value of {} was not set to {} within timeout of {} ms".format(attribute, value, retry_timeout))
                else:
                    return True
            else:
                return success

        if callback is None:
            return _wait()
        else:
            thread = Thread(target=_wait, args=[callback])
            thread.start()
            return thread

    def is_alive(self):
        """
        Check if device is still connected
        :return: bool
        """
        if not self.connected or self.closing or self.response_error:
            return False
        time_diff = self.request_timestamp - self.response_timestamp
        if time_diff >= self.response_timeout * 1e-3:
            self.handle_response_error(u"Response timeout")
            return False
        return True

    @expose_method()
    def server_disconnect(self):
        """
        This function serves just as a placeholder for
        disconnect function, but visible to client.
        """
        self.disconnect()

    @expose_method()
    def server_reconnect(self):
        """
        This function serves just as a placeholder for
        reconnect function, but visible to client.
        """
        self.reconnect()
       


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'attributeSetpointHistoryTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.13.1
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(785, 266)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.button_set_value = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_set_value.sizePolicy().hasHeightForWidth())
        self.button_set_value.setSizePolicy(sizePolicy)
        self.button_set_value.setMinimumSize(QtCore.QSize(35, 35))
        self.button_set_value.setMaximumSize(QtCore.QSize(40, 40))
        self.button_set_value.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/target.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_set_value.setIcon(icon)
        self.button_set_value.setIconSize(QtCore.QSize(25, 25))
        self.button_set_value.setObjectName("button_set_value")
        self.verticalLayout.addWidget(self.button_set_value)
        self.button_refresh = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_refresh.sizePolicy().hasHeightForWidth())
        self.button_refresh.setSizePolicy(sizePolicy)
        self.button_refresh.setMinimumSize(QtCore.QSize(35, 35))
        self.button_refresh.setMaximumSize(QtCore.QSize(40, 40))
        self.button_refresh.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/reset_bw.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_refresh.setIcon(icon1)
        self.button_refresh.setIconSize(QtCore.QSize(25, 25))
        self.button_refresh.setObjectName("button_refresh")
        self.verticalLayout.addWidget(self.button_refresh)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.button_reset = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_reset.sizePolicy().hasHeightForWidth())
        self.button_reset.setSizePolicy(sizePolicy)
        self.button_reset.setMinimumSize(QtCore.QSize(35, 35))
        self.button_reset.setMaximumSize(QtCore.QSize(40, 40))
        self.button_reset.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/cancel.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_reset.setIcon(icon2)
        self.button_reset.setIconSize(QtCore.QSize(25, 25))
        self.button_reset.setObjectName("button_reset")
        self.verticalLayout.addWidget(self.button_reset)
        self.gridLayout.addLayout(self.verticalLayout, 0, 1, 1, 1)
        self.table_history_view = QtWidgets.QTableView(Form)
        self.table_history_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.table_history_view.setObjectName("table_history_view")
        self.table_history_view.horizontalHeader().setStretchLastSection(True)
        self.gridLayout.addWidget(self.table_history_view, 0, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_set_value.clicked.connect(Form.set_attribute_value)
        self.button_refresh.clicked.connect(Form.refresh)
        self.button_reset.clicked.connect(Form.reset)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.button_set_value.setToolTip(_translate("Form", "Set attribute to saved value"))
        self.button_refresh.setToolTip(_translate("Form", "Refresh table of setpoints"))
        self.button_reset.setToolTip(_translate("Form", "Reset all setpoints"))
from kamzik3 import resource_kamzik3_rc



import time

import numpy as np
import pandas

from kamzik3.constants import STATUS_CONFIGURED
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceLogParser(Device):

    def __init__(self, device_id=None, config=None):
        Device.__init__(self, device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    @expose_method({"path": "Path to the log file", "format": "Name of the format"})
    def parse(self, path, format, filter=None):
        if format == "mag_log":
            dtypes = {}
            header_rows = []
            data_frames = []
            with open(path, "r") as f:
                for row, line in enumerate(f):
                    if line[0] == "#":
                        header_rows.append((row, line))

            for index, (header_row, header) in enumerate(header_rows):
                try:
                    start_row, end_row = header_row, header_rows[index + 1][0] - 1
                except IndexError:
                    start_row, end_row = header_row, row

                nrows = (end_row - start_row)
                if nrows <= 0:
                    continue

                headers, types = [], []
                raw_headers = header.strip().split(";")
                for raw_header in raw_headers:
                    header, meta = raw_header.split(" (")
                    dtype, unit = meta.split(",")
                    try:
                        dtype = np.dtype(dtype).name
                        if dtype in ("int64", "bool"):
                            dtype = "float64"
                    except TypeError:
                        dtype = "str"

                    header_name = "{} ({})".format(header, unit[:-1])
                    headers.append(header_name)
                    dtypes[header_name] = dtype

                df = pandas.read_csv(path, sep=";", low_memory=True, skiprows=start_row, nrows=nrows,
                                     header=0, names=headers, dtype=dtypes)
                data_frames.append(df)

        data_frame = pandas.concat(data_frames, copy=True)

        if filter is not None:
            for filter_column, filter_range in filter.items():
                data_frame = data_frame[(data_frame[filter_column] >= float(filter_range[0])) & (
                        data_frame[filter_column] <= float(filter_range[1]))]

        return data_frame



from threading import Thread, Event


class CallbackTimerInterface(Thread):

    timeout = 0

    def __init__(self, timeout, callback, with_correction=True):
        """
        Callback timer
        :param timeout: int
        :arg timeout: milliseconds
        :param callback: fun
        """
        assert callable(callback)
        self.callback = callback
        self.set_timeout(timeout)
        self.stopped = Event()
        self.with_correction = with_correction
        Thread.__init__(self)

    def set_timeout(self, timeout):
        self.timeout = timeout * 1e-3

    def run(self):
        raise NotImplementedError

    def stop(self):
        try:
            self.stopped.set()
            # self.join()
        except TypeError:
            pass